﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2010 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Security.Permissions;
using HeuristicLab.Hive.Contracts;
using HeuristicLab.Hive.Contracts.BusinessObjects;
using HeuristicLab.Hive.Contracts.Interfaces;
using HeuristicLab.Hive.Contracts.ResponseObjects;
using HeuristicLab.Hive.Server.DataAccess;

namespace HeuristicLab.Hive.Server.Core {
  public class ClientFacade : IClientFacade {

    private IJobManager jobManager = ServiceLocator.GetJobManager();
    private IContextFactory contextFactory = ServiceLocator.GetContextFactory();

    public ClientFacade() {
    }

    #region IClientFacade Members

    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.User)]
    public ResponseObject<JobDto> AddJob(SerializedJob job) {
      using (contextFactory.GetContext()) {
        return jobManager.AddNewJob(job);
      }
    }

    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.User)]
    public ResponseObject<JobDto> AddJobWithGroupStrings(SerializedJob job, IEnumerable<string> resources) {
      using (contextFactory.GetContext()) {
        return jobManager.AddJobWithGroupStrings(job, resources);
      }
    }

    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.User)]
    public Response RequestSnapshot(Guid jobId) {
      using (contextFactory.GetContext()) {
        ServiceLocator.GetAuthorizationManager().AuthorizeJobs(jobId);
        return jobManager.RequestSnapshot(jobId);
      }
    }

    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.User)]
    public ResponseObject<SerializedJob> GetLastSerializedResult(Guid jobId) {
      using (contextFactory.GetContext(false)) {
        ServiceLocator.GetAuthorizationManager().AuthorizeJobs(jobId);
        return jobManager.GetLastSerializedResult(jobId);
      }
    }

    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.User)]
    public ResponseObject<SerializedJob> GetSnapshotResult(Guid jobId) {
      using (contextFactory.GetContext(false)) {
        ServiceLocator.GetAuthorizationManager().AuthorizeJobs(jobId);
        return jobManager.GetSnapshotResult(jobId);
      }
    }

    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.User)]
    public ResponseObject<JobDto> GetJobById(Guid jobId) {
      using (contextFactory.GetContext(false)) {
        ServiceLocator.GetAuthorizationManager().AuthorizeJobs(jobId);
        return jobManager.GetJobById(jobId);
      }
    }

    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.User)]
    public Response AbortJob(Guid jobId) {
      using (contextFactory.GetContext()) {
        ServiceLocator.GetAuthorizationManager().AuthorizeJobs(jobId);
        return jobManager.AbortJob(jobId);
      }
    }

    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.User)]
    public ResponseObject<JobResultList> GetJobResults(IEnumerable<Guid> jobIds) {
      using (contextFactory.GetContext(false)) {
        ServiceLocator.GetAuthorizationManager().AuthorizeJobs(jobIds.ToArray());
        return jobManager.GetJobResults(jobIds);
      }
    }

    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.User)]
    public ResponseObject<JobResultList> GetChildJobResults(Guid? parentJobId, bool recursive, bool includeParent) {
      using (contextFactory.GetContext(false)) {
        if (parentJobId.HasValue) {
          ServiceLocator.GetAuthorizationManager().AuthorizeJobs(parentJobId.Value);
        } 
        // If parentJobId is null, GetChildJobResults will filter the results for the current user
        return jobManager.GetChildJobResults(parentJobId, recursive, includeParent);
      }
    }

    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.User)]
    public ResponseObject<HiveExperimentDtoList> GetHiveExperiments() {
      using (contextFactory.GetContext(false)) {
        return jobManager.GetHiveExperiments();
      }
    }

    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.User)]
    public ResponseObject<HiveExperimentDto> UpdateHiveExperiment(HiveExperimentDto hiveExperimentDto) {
      using (contextFactory.GetContext()) {
        if (hiveExperimentDto.UserId == Guid.Empty) {
          hiveExperimentDto.UserId = ServiceLocator.GetAuthorizationManager().UserId;
        } else {
          ServiceLocator.GetAuthorizationManager().Authorize(hiveExperimentDto.UserId);
        }
        return jobManager.UpdateHiveExperiment(hiveExperimentDto);
      }
    }

    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.User)]
    public Response DeleteHiveExperiment(Guid hiveExperimentId) {
      using (contextFactory.GetContext()) {
        var hiveExperimentDto = DaoLocator.HiveExperimentDao.FindById(hiveExperimentId);
        if (hiveExperimentDto != null) {
          ServiceLocator.GetAuthorizationManager().Authorize(hiveExperimentDto.UserId);
        }
        return jobManager.DeleteHiveExperiment(hiveExperimentId);
      }
    }

    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.User)]
    public Response DeleteJob(Guid jobId) {
      using (contextFactory.GetContext()) {
        var jobDto = DaoLocator.JobDao.FindById(jobId);
        if (jobId != null) {
          ServiceLocator.GetAuthorizationManager().Authorize(jobDto.UserId);
        }
        return jobManager.DeleteJob(jobId);
      }
    }

    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.Administrator)]
    [PrincipalPermission(SecurityAction.Demand, Role = HiveRoles.User)]
    public ResponseObject<JobDto> AddChildJob(Guid parentJobId, SerializedJob serializedJob) {
      using (contextFactory.GetContext()) {
        ServiceLocator.GetAuthorizationManager().AuthorizeJobs(parentJobId);
        return jobManager.AddChildJob(parentJobId, serializedJob);
      }
    }
    #endregion
  }
}
